using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.XR.ARFoundation;

namespace SpellBoundAR.ARImageTracking
{
    [RequireComponent(typeof(ARTrackedImageManager))]
    public class MultipleImageTracking : MonoBehaviour
    {
        public static event Action<ARTrackedImage, ImageTargetData> OnAnyTargetTracked;
        public static event Action<ARTrackedImage, ImageTargetData> OnAnyTargetRemoved;

        [Header("Settings")]
        public bool instantiateContent = true;
        [SerializeField] private Database database;

        [Header("Cache")]
        private ARTrackedImageManager _trackedImageManager;
        private readonly Dictionary<ARTrackedImage, Tuple<ImageTargetData, GameObject>> _content = new();

        public Database Database => database;
        public Dictionary<ARTrackedImage, Tuple<ImageTargetData, GameObject>> Content => _content;

        private void Awake() => _trackedImageManager = GetComponent<ARTrackedImageManager>();
        private void OnEnable() => _trackedImageManager.trackedImagesChanged += ImageStatusChanged;
        private void OnDisable() => _trackedImageManager.trackedImagesChanged -= ImageStatusChanged;

        private void ImageStatusChanged(ARTrackedImagesChangedEventArgs status)
        {
            foreach (ARTrackedImage image in status.added) AddContent(image);
            foreach (ARTrackedImage image in status.updated) UpdateContent(image);
            foreach (ARTrackedImage image in status.removed) RemoveContent(image);
        }

        private void AddContent(ARTrackedImage image)
        {
            if (Content.ContainsKey(image)) return;

            ImageTargetData imageTargetData = database ? database.GetTargetData(image) : null;
            if (!imageTargetData) return;

            if (!Content.ContainsKey(image))
            {
                Content.Add(image, null);
            }
            
            if (instantiateContent && imageTargetData.ContentEnabled)
            {
                GameObject contentObject = null;
                GameObject contentPrefab = imageTargetData.Content;
                if (contentPrefab) contentObject = Instantiate(contentPrefab);
                if (contentObject) Content[image] = new Tuple<ImageTargetData, GameObject>(imageTargetData, contentObject);
            }
        
            if (Content[image] != null)
            {
                Transform imageTransform = image.transform;
                Content[image].Item2.transform.position = imageTransform.position + Content[image].Item1.PositionOffset;
                Content[image].Item2.transform.rotation = imageTransform.rotation * Content[image].Item1.RotationOffset;  
            }
            
            imageTargetData.OnTracked();
            OnAnyTargetTracked?.Invoke(image, imageTargetData);
        }

        private void UpdateContent(ARTrackedImage image)
        {
            switch (image.trackingState) 
            {
                case UnityEngine.XR.ARSubsystems.TrackingState.Tracking:
                    AddContent(image);
                    if (Content[image] != null)
                    {
                        Transform imageTransform = image.transform;
                        Content[image].Item2.transform.position = imageTransform.position + Content[image].Item1.PositionOffset;
                        Content[image].Item2.transform.rotation = imageTransform.rotation * Content[image].Item1.RotationOffset;
                    }
                    break;
                case UnityEngine.XR.ARSubsystems.TrackingState.Limited:
                case UnityEngine.XR.ARSubsystems.TrackingState.None:
                    RemoveContent(image);
                    break;
            }
        }

        private void RemoveContent(ARTrackedImage image)
        {
            if (!Content.ContainsKey(image)) return;
            
            if (Content[image] != null)
            {
                Destroy(Content[image].Item2);
                Content[image] = null;
            }
            
            Content.Remove(image);
        
            Resources.UnloadUnusedAssets();
            
            ImageTargetData imageTargetData = database ? database.GetTargetData(image) : null;

            OnAnyTargetRemoved?.Invoke(image, imageTargetData);
        }
    }
}